@@ -22,7 +22,6 @@ import java.io.IOException;
 import java.io.UnsupportedEncodingException;
 import java.nio.charset.Charset;
 import java.nio.charset.StandardCharsets;
-import java.security.MessageDigest;
 import java.security.NoSuchAlgorithmException;
 import java.security.Principal;
 import java.security.cert.X509Certificate;
@@ -41,6 +40,7 @@ import org.apache.catalina.Engine;
 import org.apache.catalina.Host;
 import org.apache.catalina.LifecycleException;
 import org.apache.catalina.LifecycleState;
+import org.apache.catalina.Realm;
 import org.apache.catalina.Server;
 import org.apache.catalina.Service;
 import org.apache.catalina.Wrapper;
@@ -69,8 +69,7 @@ import org.ietf.jgss.GSSName;
  *
  * @author Craig R. McClanahan
  */
-@SuppressWarnings("deprecation")
-public abstract class RealmBase extends LifecycleMBeanBase implements org.apache.catalina.GSSRealm {
+public abstract class RealmBase extends LifecycleMBeanBase implements Realm {
 
     private static final Log log = LogFactory.getLog(RealmBase.class);
 
@@ -1052,11 +1051,6 @@ public abstract class RealmBase extends LifecycleMBeanBase implements org.apache
     }
 
 
-    @Override
-    public boolean isAvailable() {
-        return true;
-    }
-
     @Override
     protected void initInternal() throws LifecycleException {
 
@@ -1190,15 +1184,6 @@ public abstract class RealmBase extends LifecycleMBeanBase implements org.apache
     }
 
 
-    /**
-     * @return a short name for this Realm implementation, for use in log messages.
-     *
-     * @deprecated This will be removed in Tomcat 9 onwards. Use {@link Class#getSimpleName()} instead.
-     */
-    @Deprecated
-    protected abstract String getName();
-
-
     /**
      * Get the password for the specified user.
      *
@@ -1314,42 +1299,6 @@ public abstract class RealmBase extends LifecycleMBeanBase implements org.apache
 
     // --------------------------------------------------------- Static Methods
 
-    /**
-     * Digest password using the algorithm specified and convert the result to a corresponding hex string.
-     *
-     * @param credentials Password or other credentials to use in authenticating this username
-     * @param algorithm   Algorithm used to do the digest
-     * @param encoding    Character encoding of the string to digest
-     *
-     * @return The digested credentials as a hex string or the original plain text credentials if an error occurs.
-     *
-     * @deprecated Unused. This will be removed in Tomcat 9.
-     */
-    @Deprecated
-    public static final String Digest(String credentials, String algorithm, String encoding) {
-
-        try {
-            // Obtain a new message digest with "digest" encryption
-            MessageDigest md = (MessageDigest) MessageDigest.getInstance(algorithm).clone();
-
-            // encode the credentials
-            // Should use the digestEncoding, but that's not a static field
-            if (encoding == null) {
-                md.update(credentials.getBytes());
-            } else {
-                md.update(credentials.getBytes(encoding));
-            }
-
-            // Digest the credentials and return as hexadecimal
-            return (HexUtils.toHexString(md.digest()));
-        } catch (Exception ex) {
-            log.error(ex);
-            return credentials;
-        }
-
-    }
-
-
     /**
      * Generate a stored credential string for the given password and associated parameters.
      * <p>
