@@ -17,11 +17,11 @@
 package org.apache.catalina.startup;
 
 import java.io.File;
+import java.io.FileInputStream;
 import java.io.FileOutputStream;
 import java.io.IOException;
 import java.io.InputStream;
 import java.io.OutputStream;
-import java.lang.reflect.InvocationTargetException;
 import java.net.MalformedURLException;
 import java.net.URL;
 import java.net.URLConnection;
@@ -35,6 +35,7 @@ import java.util.Stack;
 import java.util.jar.JarEntry;
 import java.util.jar.JarFile;
 import java.util.logging.Level;
+import java.util.logging.LogManager;
 import java.util.logging.Logger;
 
 import javax.servlet.Servlet;
@@ -66,11 +67,16 @@ import org.apache.catalina.core.StandardService;
 import org.apache.catalina.core.StandardWrapper;
 import org.apache.catalina.realm.GenericPrincipal;
 import org.apache.catalina.realm.RealmBase;
+import org.apache.catalina.security.SecurityClassLoad;
 import org.apache.catalina.util.ContextName;
 import org.apache.catalina.util.IOTools;
 import org.apache.tomcat.util.ExceptionUtils;
 import org.apache.tomcat.util.buf.UriUtil;
+import org.apache.tomcat.util.compat.JreCompat;
 import org.apache.tomcat.util.descriptor.web.LoginConfig;
+import org.apache.tomcat.util.file.ConfigFileLoader;
+import org.apache.tomcat.util.file.ConfigurationSource;
+import org.apache.tomcat.util.modeler.Registry;
 import org.apache.tomcat.util.res.StringManager;
 
 // TODO: lazy init for the temp dir - only when a JSP is compiled or
@@ -166,7 +172,6 @@ public class Tomcat {
     protected int port = 8080;
     protected String hostname = "localhost";
     protected String basedir;
-    protected boolean defaultConnectorCreated = false;
 
     private final Map<String, String> userPass = new HashMap<>();
     private final Map<String, List<String>> userRoles = new HashMap<>();
@@ -179,30 +184,34 @@ public class Tomcat {
     }
 
     /**
-     * Tomcat needs a directory for temp files. This should be the
-     * first method called.
-     *
+     * Tomcat requires that the base directory is set because the defaults for
+     * a number of other locations, such as the work directory, are derived from
+     * the base directory. This should be the first method called.
      * <p>
-     * By default, if this method is not called, we use:
-     * <ul>
-     *  <li>system properties - catalina.base, catalina.home</li>
-     *  <li>[user.dir]/tomcat.$PORT</li>
-     * </ul>
-     * (/tmp doesn't seem a good choice for security).
-     *
+     * If this method is not called then Tomcat will attempt to use these
+     * locations in the following order:
+     * <ol>
+     *  <li>if set, the catalina.base system property</li>
+     *  <li>if set, the catalina.home system property</li>
+     *  <li>The user.dir system property (the directory where Java was run from)
+     *      where a directory named tomcat.$PORT will be created. $PORT is the
+     *      value configured via {@link #setPort(int)} which defaults to 8080 if
+     *      not set</li>
+     * </ol>
+     * The user should ensure that the file permissions for the base directory
+     * are appropriate.
      * <p>
      * TODO: disable work dir if not needed ( no jsp, etc ).
      *
-     * @param basedir The Tomcat base folder on which all others
-     *  will be derived
+     * @param basedir The Tomcat base folder on which all others will be derived
      */
     public void setBaseDir(String basedir) {
         this.basedir = basedir;
     }
 
     /**
-     * Set the port for the default connector. Must
-     * be called before start().
+     * Set the port for the default connector. The default connector will
+     * only be created if getConnector is called.
      * @param port The port number
      */
     public void setPort(int port) {
@@ -376,6 +385,9 @@ public class Tomcat {
                                       String servletClass) {
         // will do class for name and set init params
         Wrapper sw = ctx.createWrapper();
+        if (sw == null) {
+            return null;
+        }
         sw.setServletClass(servletClass);
         sw.setName(servletName);
         ctx.addChild(sw);
@@ -417,6 +429,45 @@ public class Tomcat {
     }
 
 
+    /**
+     * Initialize the server given the specified configuration source.
+     * The server will be loaded according to the Tomcat configuration
+     * files contained in the source (server.xml, web.xml, context.xml,
+     * SSL certificates, etc).
+     * If no configuration source is specified, it will use the default
+     * locations for these files.
+     * @param source The configuration source
+     */
+    public void init(ConfigurationSource source) {
+        init(source, null);
+    }
+
+    /**
+     * Initialize the server given the specified configuration source.
+     * The server will be loaded according to the Tomcat configuration
+     * files contained in the source (server.xml, web.xml, context.xml,
+     * SSL certificates, etc).
+     * If no configuration source is specified, it will use the default
+     * locations for these files.
+     * @param source The configuration source
+     * @param catalinaArguments The arguments that should be passed to Catalina
+     */
+    public void init(ConfigurationSource source, String[] catalinaArguments) {
+        ConfigFileLoader.setSource(source);
+        addDefaultWebXmlToWebapp = false;
+        Catalina catalina = new Catalina();
+        // Load the Catalina instance with the regular configuration files
+        // from specified source
+        if (catalinaArguments == null) {
+            catalina.load();
+        } else {
+            catalina.load(catalinaArguments);
+        }
+        // Retrieve and set the server
+        server = catalina.getServer();
+    }
+
+
     /**
      * Initialize the server.
      *
@@ -424,7 +475,6 @@ public class Tomcat {
      */
     public void init() throws LifecycleException {
         getServer();
-        getConnector();
         server.init();
     }
 
@@ -436,7 +486,6 @@ public class Tomcat {
      */
     public void start() throws LifecycleException {
         getServer();
-        getConnector();
         server.start();
     }
 
@@ -480,36 +529,26 @@ public class Tomcat {
      * @param role The role name
      */
     public void addRole(String user, String role) {
-        List<String> roles = userRoles.get(user);
-        if (roles == null) {
-            roles = new ArrayList<>();
-            userRoles.put(user, roles);
-        }
-        roles.add(role);
+        userRoles.computeIfAbsent(user, k -> new ArrayList<>()).add(role);
     }
 
     // ------- Extra customization -------
     // You can tune individual Tomcat objects, using internal APIs
 
     /**
-     * Get the default http connector. You can set more
-     * parameters - the port is already initialized.
+     * Get the default HTTP connector that is used by the embedded
+     * Tomcat. It is first configured connector in the service.
+     * If there's no connector defined, it will create and add a default
+     * connector using the port and address specified in this Tomcat
+     * instance, and return it for further customization.
      *
-     * <p>
-     * Alternatively, you can construct a Connector and set any params,
-     * then call addConnector(Connector)
-     *
-     * @return A connector object that can be customized
+     * @return The connector object
      */
     public Connector getConnector() {
         Service service = getService();
         if (service.findConnectors().length > 0) {
             return service.findConnectors()[0];
         }
-
-        if (defaultConnectorCreated) {
-            return null;
-        }
         // The same as in standard Tomcat configuration.
         // This creates an APR HTTP connector if AprLifecycleListener has been
         // configured (created) and Tomcat Native library is available.
@@ -517,12 +556,15 @@ public class Tomcat {
         Connector connector = new Connector("HTTP/1.1");
         connector.setPort(port);
         service.addConnector(connector);
-        defaultConnectorCreated = true;
         return connector;
     }
 
+    /**
+     * Set the specified connector in the service, if it is not already
+     * present.
+     * @param connector The connector instance to add
+     */
     public void setConnector(Connector connector) {
-        defaultConnectorCreated = true;
         Service service = getService();
         boolean found = false;
         for (Connector serviceConnector : service.findConnectors()) {
@@ -612,6 +654,9 @@ public class Tomcat {
 
         initBaseDir();
 
+        // Set configuration source
+        ConfigFileLoader.setSource(new CatalinaBaseConfigurationSource(new File(basedir), null));
+
         server.setPort( -1 );
 
         Service service = new StandardService();
@@ -692,24 +737,6 @@ public class Tomcat {
     }
 
 
-    /**
-     * @param host The host in which the context will be deployed
-     * @param contextPath The context mapping to use, "" for root context.
-     * @param docBase Base directory for the context, for static files.
-     *  Must exist, relative to the server home
-     * @param config Custom context configurator helper
-     * @return the deployed context
-     * @see #addWebapp(String, String)
-     *
-     * @deprecated Use {@link
-     *             #addWebapp(Host, String, String, LifecycleListener)} instead
-     */
-    @Deprecated
-    public Context addWebapp(Host host, String contextPath, String docBase, ContextConfig config) {
-        return addWebapp(host, contextPath, docBase, (LifecycleListener) config);
-    }
-
-
     /**
      * This is equivalent to adding a web application to a Host's appBase
      * (usually Tomcat's webapps directory). By default, the equivalent of the
@@ -793,33 +820,30 @@ public class Tomcat {
      * @return a realm instance
      */
     protected Realm createDefaultRealm() {
-        return new RealmBase() {
-            @Override
-            @Deprecated
-            protected String getName() {
-                return "Simple";
-            }
+        return new SimpleRealm();
+    }
 
-            @Override
-            protected String getPassword(String username) {
-                return userPass.get(username);
-            }
 
-            @Override
-            protected Principal getPrincipal(String username) {
-                Principal p = userPrincipals.get(username);
-                if (p == null) {
-                    String pass = userPass.get(username);
-                    if (pass != null) {
-                        p = new GenericPrincipal(username, pass,
-                                userRoles.get(username));
-                        userPrincipals.put(username, p);
-                    }
+    private class SimpleRealm extends RealmBase {
+
+        @Override
+        protected String getPassword(String username) {
+            return userPass.get(username);
+        }
+
+        @Override
+        protected Principal getPrincipal(String username) {
+            Principal p = userPrincipals.get(username);
+            if (p == null) {
+                String pass = userPass.get(username);
+                if (pass != null) {
+                    p = new GenericPrincipal(username, pass,
+                            userRoles.get(username));
+                    userPrincipals.put(username, p);
                 }
-                return p;
             }
-
-        };
+            return p;
+        }
     }
 
 
@@ -985,6 +1009,7 @@ public class Tomcat {
      * @return newly created {@link Context}
      */
     private Context createContext(Host host, String url) {
+        String defaultContextClass = StandardContext.class.getName();
         String contextClass = StandardContext.class.getName();
         if (host == null) {
             host = this.getHost();
@@ -993,16 +1018,15 @@ public class Tomcat {
             contextClass = ((StandardHost) host).getContextClass();
         }
         try {
-            return (Context) Class.forName(contextClass).getConstructor()
+            if (defaultContextClass.equals(contextClass)) {
+                return new StandardContext();
+            } else {
+                return (Context) Class.forName(contextClass).getConstructor()
                     .newInstance();
-        } catch (InstantiationException | IllegalAccessException
-                | IllegalArgumentException | InvocationTargetException
-                | NoSuchMethodException | SecurityException
-                | ClassNotFoundException e) {
-            throw new IllegalArgumentException(
-                    "Can't instantiate context-class " + contextClass
-                            + " for host " + host + " and url "
-                            + url, e);
+            }
+
+        } catch (ReflectiveOperationException  | IllegalArgumentException | SecurityException e) {
+            throw new IllegalArgumentException(sm.getString("tomcat.noContextClass", contextClass, host, url), e);
         }
     }
 
@@ -1136,8 +1160,11 @@ public class Tomcat {
                 if (event.getType().equals(Lifecycle.CONFIGURE_START_EVENT)) {
                     context.setConfigured(true);
 
-                    // Process annotations
-                    WebAnnotationSet.loadApplicationAnnotations(context);
+                    // Process annotations when not running in a Graal image
+                    // annotations require reflections and additional configuration
+                    if (!JreCompat.isGraalAvailable()) {
+                        WebAnnotationSet.loadApplicationAnnotations(context);
+                    }
 
                     // LoginConfig is required to process @ServletSecurity
                     // annotations
@@ -1250,7 +1277,7 @@ public class Tomcat {
                 result = webAppContextXml.toURI().toURL();
             } catch (MalformedURLException e) {
                 Logger.getLogger(getLoggerName(getHost(), contextName)).log(Level.WARNING,
-                        "Unable to determine web application context.xml " + docBase, e);
+                        sm.getString("tomcat.noContextXml", docBase), e);
             }
         }
         return result;
@@ -1265,8 +1292,82 @@ public class Tomcat {
             }
         } catch (IOException e) {
             Logger.getLogger(getLoggerName(getHost(), contextName)).log(Level.WARNING,
-                    "Unable to determine web application context.xml " + docBase, e);
+                    sm.getString("tomcat.noContextXml", docBase), e);
         }
         return result;
     }
+
+    static {
+        // Graal native images don't load any configuration except the VM default
+        if (JreCompat.isGraalAvailable()) {
+            try (InputStream is = new FileInputStream(new File(System.getProperty("java.util.logging.config.file", "conf/logging.properties")))) {
+                LogManager.getLogManager().readConfiguration(is);
+            } catch (SecurityException | IOException e) {
+                // Ignore, the VM default will be used
+            }
+        }
+    }
+
+    /**
+     * Main executable method for use with a Maven packager.
+     * @param args the command line arguments
+     * @throws Exception if an error occurs
+     */
+    public static void main(String[] args) throws Exception {
+        // Process some command line parameters
+        String[] catalinaArguments = null;
+        for (int i = 0; i < args.length; i++) {
+            if (args[i].equals("--no-jmx")) {
+                Registry.disableRegistry();
+            } else if (args[i].equals("--catalina")) {
+                // This was already processed before
+                // Skip the rest of the arguments as they are for Catalina
+                ArrayList<String> result = new ArrayList<>();
+                for (int j = i + 1; j < args.length; j++) {
+                    result.add(args[j]);
+                }
+                catalinaArguments = result.toArray(new String[0]);
+                break;
+            }
+        }
+        SecurityClassLoad.securityClassLoad(Thread.currentThread().getContextClassLoader());
+        Tomcat tomcat = new Tomcat();
+        // Create a Catalina instance and let it parse the configuration files
+        // It will also set a shutdown hook to stop the Server when needed
+        // Use the default configuration source
+        tomcat.init(null, catalinaArguments);
+        boolean await = false;
+        String path = "";
+        // Process command line parameters
+        for (int i = 0; i < args.length; i++) {
+            if (args[i].equals("--war")) {
+                if (++i >= args.length) {
+                    throw new IllegalArgumentException(sm.getString("tomcat.invalidCommandLine", args[i - 1]));
+                }
+                File war = new File(args[i]);
+                tomcat.addWebapp(path, war.getAbsolutePath());
+            } else if (args[i].equals("--path")) {
+                if (++i >= args.length) {
+                    throw new IllegalArgumentException(sm.getString("tomcat.invalidCommandLine", args[i - 1]));
+                }
+                path = args[i];
+            } else if (args[i].equals("--await")) {
+                await = true;
+            } else if (args[i].equals("--no-jmx")) {
+                // This was already processed before
+            } else if (args[i].equals("--catalina")) {
+                // This was already processed before
+                // Skip the rest of the arguments as they are for Catalina
+                break;
+            } else {
+                throw new IllegalArgumentException(sm.getString("tomcat.invalidCommandLine", args[i]));
+            }
+        }
+        tomcat.start();
+        // Ideally the utility threads are non daemon
+        if (await) {
+            tomcat.getServer().await();
+        }
+    }
+
 }
