@@ -26,6 +26,8 @@ import java.util.ArrayList;
 import java.util.List;
 import java.util.concurrent.TimeUnit;
 import java.util.concurrent.atomic.AtomicReference;
+import java.util.concurrent.locks.Lock;
+import java.util.concurrent.locks.ReentrantLock;
 
 import javax.servlet.http.WebConnection;
 
@@ -44,9 +46,9 @@ public class Http2AsyncUpgradeHandler extends Http2UpgradeHandler {
     // Ensures headers are generated and then written for one thread at a time.
     // Because of the compression used, headers need to be written to the
     // network in the same order they are generated.
-    private final Object headerWriteLock = new Object();
+    private final Lock headerWriteLock = new ReentrantLock();
     // Ensures thread triggers the stream reset is the first to send a RST frame
-    private final Object sendResetLock = new Object();
+    private final Lock sendResetLock = new ReentrantLock();
     private final AtomicReference<Throwable> error = new AtomicReference<>();
     private final AtomicReference<IOException> applicationIOE = new AtomicReference<>();
 
@@ -54,7 +56,7 @@ public class Http2AsyncUpgradeHandler extends Http2UpgradeHandler {
         super(protocol, adapter, coyoteRequest);
     }
 
-    private final CompletionHandler<Long, Void> errorCompletion = new CompletionHandler<Long, Void>() {
+    private final CompletionHandler<Long,Void> errorCompletion = new CompletionHandler<Long,Void>() {
         @Override
         public void completed(Long result, Void attachment) {
         }
@@ -64,7 +66,7 @@ public class Http2AsyncUpgradeHandler extends Http2UpgradeHandler {
             error.set(t);
         }
     };
-    private final CompletionHandler<Long, Void> applicationErrorCompletion = new CompletionHandler<Long, Void>() {
+    private final CompletionHandler<Long,Void> applicationErrorCompletion = new CompletionHandler<Long,Void>() {
         @Override
         public void completed(Long result, Void attachment) {
         }
@@ -148,7 +150,8 @@ public class Http2AsyncUpgradeHandler extends Http2UpgradeHandler {
         // may see out of order RST frames which may hard to follow if
         // the client is unaware the RST frames may be received out of
         // order.
-        synchronized (sendResetLock) {
+        sendResetLock.lock();
+        try {
             if (state != null) {
                 boolean active = state.isActive();
                 state.sendReset();
@@ -159,6 +162,8 @@ public class Http2AsyncUpgradeHandler extends Http2UpgradeHandler {
 
             socketWrapper.write(BlockingMode.SEMI_BLOCK, protocol.getWriteTimeout(), TimeUnit.MILLISECONDS, null,
                     SocketWrapperBase.COMPLETE_WRITE, errorCompletion, ByteBuffer.wrap(rstFrame));
+        } finally {
+            sendResetLock.unlock();
         }
         handleAsyncException();
     }
@@ -191,7 +196,8 @@ public class Http2AsyncUpgradeHandler extends Http2UpgradeHandler {
     @Override
     void writeHeaders(Stream stream, int pushedStreamId, MimeHeaders mimeHeaders, boolean endOfStream, int payloadSize)
             throws IOException {
-        synchronized (headerWriteLock) {
+        headerWriteLock.lock();
+        try {
             AsyncHeaderFrameBuffers headerFrameBuffers = (AsyncHeaderFrameBuffers) doWriteHeaders(stream,
                     pushedStreamId, mimeHeaders, endOfStream, payloadSize);
             if (headerFrameBuffers != null) {
@@ -200,9 +206,11 @@ public class Http2AsyncUpgradeHandler extends Http2UpgradeHandler {
                         headerFrameBuffers.bufs.toArray(BYTEBUFFER_ARRAY));
                 handleAsyncException();
             }
+        } finally {
+            headerWriteLock.unlock();
         }
         if (endOfStream) {
-            stream.sentEndOfStream();
+            sentEndOfStream(stream);
         }
     }
 
@@ -219,6 +227,9 @@ public class Http2AsyncUpgradeHandler extends Http2UpgradeHandler {
             log.debug(sm.getString("upgradeHandler.writeBody", connectionId, stream.getIdAsString(),
                     Integer.toString(len), Boolean.valueOf(finished)));
         }
+
+        reduceOverheadCount(FrameType.DATA);
+
         // Need to check this now since sending end of stream will change this.
         boolean writable = stream.canWrite();
         byte[] header = new byte[9];
@@ -226,10 +237,7 @@ public class Http2AsyncUpgradeHandler extends Http2UpgradeHandler {
         header[3] = FrameType.DATA.getIdByte();
         if (finished) {
             header[4] = FLAG_END_OF_STREAM;
-            stream.sentEndOfStream();
-            if (!stream.isActive()) {
-                setConnectionTimeoutForStreamCount(activeRemoteStreamCount.decrementAndGet());
-            }
+            sentEndOfStream(stream);
         }
         if (writable) {
             ByteUtil.set31Bits(header, 5, stream.getIdAsInt());
@@ -322,8 +330,8 @@ public class Http2AsyncUpgradeHandler extends Http2UpgradeHandler {
                     sendfile.mappedBuffer = channel.map(MapMode.READ_ONLY, sendfile.pos, sendfile.end - sendfile.pos);
                 }
                 // Reserve as much as possible right away
-                int reservation = (sendfile.end - sendfile.pos > Integer.MAX_VALUE) ? Integer.MAX_VALUE
-                        : (int) (sendfile.end - sendfile.pos);
+                int reservation = (sendfile.end - sendfile.pos > Integer.MAX_VALUE) ? Integer.MAX_VALUE :
+                        (int) (sendfile.end - sendfile.pos);
                 sendfile.streamReservation = sendfile.stream.reserveWindowSize(reservation, true);
                 sendfile.connectionReservation = reserveWindowSize(sendfile.stream, sendfile.streamReservation, true);
             } catch (IOException e) {
@@ -339,8 +347,8 @@ public class Http2AsyncUpgradeHandler extends Http2UpgradeHandler {
             // connectionReservation will always be smaller than or the same as
             // streamReservation
             int frameSize = Integer.min(getMaxFrameSize(), sendfile.connectionReservation);
-            boolean finished = (frameSize == sendfile.left) &&
-                    sendfile.stream.getCoyoteResponse().getTrailerFields() == null;
+            boolean finished =
+                    (frameSize == sendfile.left) && sendfile.stream.getCoyoteResponse().getTrailerFields() == null;
 
             // Need to check this now since sending end of stream will change this.
             boolean writable = sendfile.stream.canWrite();
@@ -349,10 +357,7 @@ public class Http2AsyncUpgradeHandler extends Http2UpgradeHandler {
             header[3] = FrameType.DATA.getIdByte();
             if (finished) {
                 header[4] = FLAG_END_OF_STREAM;
-                sendfile.stream.sentEndOfStream();
-                if (!sendfile.stream.isActive()) {
-                    setConnectionTimeoutForStreamCount(activeRemoteStreamCount.decrementAndGet());
-                }
+                sentEndOfStream(sendfile.stream);
             }
             if (writable) {
                 if (log.isDebugEnabled()) {
@@ -376,7 +381,7 @@ public class Http2AsyncUpgradeHandler extends Http2UpgradeHandler {
         }
     }
 
-    protected class SendfileCompletionHandler implements CompletionHandler<Long, SendfileData> {
+    protected class SendfileCompletionHandler implements CompletionHandler<Long,SendfileData> {
         @Override
         public void completed(Long nBytes, SendfileData sendfile) {
             CompletionState completionState = null;
@@ -402,12 +407,12 @@ public class Http2AsyncUpgradeHandler extends Http2UpgradeHandler {
                 try {
                     if (sendfile.connectionReservation == 0) {
                         if (sendfile.streamReservation == 0) {
-                            int reservation = (sendfile.end - sendfile.pos > Integer.MAX_VALUE) ? Integer.MAX_VALUE
-                                    : (int) (sendfile.end - sendfile.pos);
+                            int reservation = (sendfile.end - sendfile.pos > Integer.MAX_VALUE) ? Integer.MAX_VALUE :
+                                    (int) (sendfile.end - sendfile.pos);
                             sendfile.streamReservation = sendfile.stream.reserveWindowSize(reservation, true);
                         }
-                        sendfile.connectionReservation = reserveWindowSize(sendfile.stream, sendfile.streamReservation,
-                                true);
+                        sendfile.connectionReservation =
+                                reserveWindowSize(sendfile.stream, sendfile.streamReservation, true);
                     }
                 } catch (IOException e) {
                     failed(e, sendfile);
@@ -423,8 +428,8 @@ public class Http2AsyncUpgradeHandler extends Http2UpgradeHandler {
                 // connectionReservation will always be smaller than or the same as
                 // streamReservation
                 int frameSize = Integer.min(getMaxFrameSize(), sendfile.connectionReservation);
-                boolean finished = (frameSize == sendfile.left) &&
-                        sendfile.stream.getCoyoteResponse().getTrailerFields() == null;
+                boolean finished =
+                        (frameSize == sendfile.left) && sendfile.stream.getCoyoteResponse().getTrailerFields() == null;
 
                 // Need to check this now since sending end of stream will change this.
                 boolean writable = sendfile.stream.canWrite();
@@ -433,10 +438,7 @@ public class Http2AsyncUpgradeHandler extends Http2UpgradeHandler {
                 header[3] = FrameType.DATA.getIdByte();
                 if (finished) {
                     header[4] = FLAG_END_OF_STREAM;
-                    sendfile.stream.sentEndOfStream();
-                    if (!sendfile.stream.isActive()) {
-                        setConnectionTimeoutForStreamCount(activeRemoteStreamCount.decrementAndGet());
-                    }
+                    sentEndOfStream(sendfile.stream);
                 }
                 if (writable) {
                     if (log.isDebugEnabled()) {
