@@ -526,19 +526,19 @@ public final class MessageBytes implements Cloneable, Serializable {
      */
     public void duplicate(MessageBytes src) throws IOException {
         switch (src.getType()) {
-            case MessageBytes.T_BYTES:
+            case T_BYTES:
                 type = T_BYTES;
                 ByteChunk bc = src.getByteChunk();
                 byteC.allocate(2 * bc.getLength(), -1);
                 byteC.append(bc);
                 break;
-            case MessageBytes.T_CHARS:
+            case T_CHARS:
                 type = T_CHARS;
                 CharChunk cc = src.getCharChunk();
                 charC.allocate(2 * cc.getLength(), -1);
                 charC.append(cc);
                 break;
-            case MessageBytes.T_STR:
+            case T_STR:
                 type = T_STR;
                 String sc = src.getString();
                 this.setString(sc);
@@ -598,7 +598,7 @@ public final class MessageBytes implements Cloneable, Serializable {
 
     // Used for headers conversion
     /**
-     * Convert the buffer to an long, cache the value.
+     * Convert the buffer to a long, cache the value.
      *
      * @return the long value
      */
