@@ -38,6 +38,8 @@ import java.util.concurrent.ConcurrentHashMap;
 import java.util.concurrent.ConcurrentMap;
 import java.util.concurrent.Executor;
 import java.util.concurrent.RejectedExecutionException;
+import java.util.concurrent.ScheduledExecutorService;
+import java.util.concurrent.ScheduledThreadPoolExecutor;
 import java.util.concurrent.TimeUnit;
 
 import javax.management.MalformedObjectNameException;
@@ -82,7 +84,7 @@ public abstract class AbstractEndpoint<S,U> {
         enum SocketState {
             // TODO Add a new state to the AsyncStateMachine and remove
             //      ASYNC_END (if possible)
-            OPEN, CLOSED, LONG, ASYNC_END, SENDFILE, UPGRADING, UPGRADED, SUSPENDED
+            OPEN, CLOSED, LONG, ASYNC_END, SENDFILE, UPGRADING, UPGRADED, ASYNC_IO, SUSPENDED
         }
 
 
@@ -284,9 +286,6 @@ public abstract class AbstractEndpoint<S,U> {
         }
         if (replace) {
             SSLHostConfig previous = sslHostConfigs.put(key, sslHostConfig);
-            if (key.equals(getDefaultSSLHostConfigName())) {
-                setDefaultSslHostConfig(sslHostConfig);
-            }
             if (previous != null) {
                 unregisterJmx(sslHostConfig);
             }
@@ -377,9 +376,6 @@ public abstract class AbstractEndpoint<S,U> {
     protected abstract void createSSLContext(SSLHostConfig sslHostConfig) throws Exception;
 
 
-    protected abstract void setDefaultSslHostConfig(SSLHostConfig sslHostConfig);
-
-
     protected void logCertificate(SSLHostConfigCertificate certificate) {
         SSLHostConfig sslHostConfig = certificate.getSSLHostConfig();
 
@@ -472,8 +468,7 @@ public abstract class AbstractEndpoint<S,U> {
     protected void releaseSSLContext(SSLHostConfig sslHostConfig) {
         for (SSLHostConfigCertificate certificate : sslHostConfig.getCertificates(true)) {
             if (certificate.getSslContext() != null) {
-                // Only release the SSLContext if we generated it.
-                SSLContext sslContext = certificate.getSslContextGenerated();
+                SSLContext sslContext = certificate.getSslContext();
                 if (sslContext != null) {
                     sslContext.destroy();
                 }
@@ -655,10 +650,26 @@ public abstract class AbstractEndpoint<S,U> {
     }
 
 
+    /**
+     * External Executor based thread pool for utility tasks.
+     */
+    private ScheduledExecutorService utilityExecutor = null;
+    public void setUtilityExecutor(ScheduledExecutorService utilityExecutor) {
+        this.utilityExecutor = utilityExecutor;
+    }
+    public ScheduledExecutorService getUtilityExecutor() {
+        if (utilityExecutor == null) {
+            getLog().warn(sm.getString("endpoint.warn.noUtilityExecutor"));
+            utilityExecutor = new ScheduledThreadPoolExecutor(1);
+        }
+        return utilityExecutor;
+    }
+
+
     /**
      * Server socket port.
      */
-    private int port;
+    private int port = -1;
     public int getPort() { return port; }
     public void setPort(int port ) { this.port=port; }
 
@@ -730,10 +741,6 @@ public abstract class AbstractEndpoint<S,U> {
         this.acceptCount = acceptCount;
     } }
     public int getAcceptCount() { return acceptCount; }
-    @Deprecated
-    public void setBacklog(int backlog) { setAcceptCount(backlog); }
-    @Deprecated
-    public int getBacklog() { return getAcceptCount(); }
 
     /**
      * Controls when the Endpoint binds the port. <code>true</code>, the default
@@ -786,10 +793,6 @@ public abstract class AbstractEndpoint<S,U> {
         socketProperties.setSoLingerTime(connectionLinger);
         socketProperties.setSoLingerOn(connectionLinger>=0);
     }
-    @Deprecated
-    public int getSoLinger() { return getConnectionLinger(); }
-    @Deprecated
-    public void setSoLinger(int soLinger) { setConnectionLinger(soLinger);}
 
 
     /**
@@ -799,10 +802,6 @@ public abstract class AbstractEndpoint<S,U> {
      */
     public int getConnectionTimeout() { return socketProperties.getSoTimeout(); }
     public void setConnectionTimeout(int soTimeout) { socketProperties.setSoTimeout(soTimeout); }
-    @Deprecated
-    public int getSoTimeout() { return getConnectionTimeout(); }
-    @Deprecated
-    public void setSoTimeout(int soTimeout) { setConnectionTimeout(soTimeout); }
 
     /**
      * SSL engine.
@@ -902,17 +901,6 @@ public abstract class AbstractEndpoint<S,U> {
         this.maxKeepAliveRequests = maxKeepAliveRequests;
     }
 
-    /**
-     * The maximum number of headers in a request that are allowed.
-     * 100 by default. A value of less than 0 means no limit.
-     */
-    private int maxHeaderCount = 100; // as in Apache HTTPD server
-    public int getMaxHeaderCount() {
-        return maxHeaderCount;
-    }
-    public void setMaxHeaderCount(int maxHeaderCount) {
-        this.maxHeaderCount = maxHeaderCount;
-    }
 
     /**
      * Name of the thread pool, which will be used for naming child threads.
@@ -951,6 +939,17 @@ public abstract class AbstractEndpoint<S,U> {
     protected abstract boolean getDeferAccept();
 
 
+    /**
+     * The default behavior is to identify connectors uniquely with address
+     * and port. However, certain connectors are not using that and need
+     * some other identifier, which then can be used as a replacement.
+     * @return the id
+     */
+    public String getId() {
+        return null;
+    }
+
+
     protected final List<String> negotiableProtocols = new ArrayList<>();
     public void addNegotiatedProtocol(String negotiableProtocol) {
         negotiableProtocols.add(negotiableProtocol);
@@ -1164,8 +1163,8 @@ public abstract class AbstractEndpoint<S,U> {
                 // That causes delays when running the unit tests. Explicitly enableing linger but with a timeout of
                 // zero seconds seems to fix the issue.
                 s.setSoLinger(true, 0);
-                if (getLog().isTraceEnabled()) {
-                    getLog().trace("About to unlock socket for:" + unlockAddress);
+                if (getLog().isDebugEnabled()) {
+                    getLog().debug("About to unlock socket for:" + unlockAddress);
                 }
                 s.connect(unlockAddress,utmo);
                 if (getDeferAccept()) {
@@ -1181,8 +1180,8 @@ public abstract class AbstractEndpoint<S,U> {
                             "User-Agent: Tomcat wakeup connection\r\n\r\n");
                     sw.flush();
                 }
-                if (getLog().isTraceEnabled()) {
-                    getLog().trace("Socket unlock completed for:" + unlockAddress);
+                if (getLog().isDebugEnabled()) {
+                    getLog().debug("Socket unlock completed for:" + unlockAddress);
                 }
             }
             // Wait for up to 1000ms acceptor threads to unlock. Particularly
@@ -1318,7 +1317,6 @@ public abstract class AbstractEndpoint<S,U> {
 
     public abstract void bind() throws Exception;
     public abstract void unbind() throws Exception;
-
     public abstract void startInternal() throws Exception;
     public abstract void stopInternal() throws Exception;
 
@@ -1336,7 +1334,7 @@ public abstract class AbstractEndpoint<S,U> {
     }
 
 
-    public void init() throws Exception {
+    public final void init() throws Exception {
         if (bindOnInit) {
             bindWithCleanup();
             bindState = BindState.BOUND_ON_INIT;
@@ -1490,7 +1488,7 @@ public abstract class AbstractEndpoint<S,U> {
         return connectionLimitLatch;
     }
 
-    protected void releaseConnectionLatch() {
+    private void releaseConnectionLatch() {
         LimitLatch latch = connectionLimitLatch;
         if (latch!=null) {
             latch.releaseAll();
