@@ -301,8 +301,8 @@ public class SecureNioChannel extends NioChannel {
             throw new IOException(sm.getString("channel.nio.ssl.foundHttp"));
         }
 
-        if (log.isTraceEnabled()) {
-            log.trace(sm.getString("channel.nio.ssl.sniHostName", sc, hostName));
+        if (log.isDebugEnabled()) {
+            log.debug(sm.getString("channel.nio.ssl.sniHostName", sc, hostName));
         }
 
         sslEngine = endpoint.createSSLEngine(hostName, clientRequestedCiphers,
@@ -463,6 +463,10 @@ public class SecureNioChannel extends NioChannel {
      */
     protected SSLEngineResult handshakeUnwrap(boolean doread) throws IOException {
 
+        if (netInBuffer.position() == netInBuffer.limit()) {
+            //clear the buffer if we have emptied it out on data
+            netInBuffer.clear();
+        }
         if (doread)  {
             //if we have data to read, read it
             int read = sc.read(netInBuffer);
@@ -479,12 +483,7 @@ public class SecureNioChannel extends NioChannel {
             //call unwrap
             getBufHandler().configureReadBufferForWrite();
             result = sslEngine.unwrap(netInBuffer, getBufHandler().getReadBuffer());
-            /*
-             * ByteBuffer.compact() is an optional method but netInBuffer is created from either ByteBuffer.allocate()
-             * or ByteBuffer.allocateDirect() and the ByteBuffers returned by those methods do implement compact().
-             * The ByteBuffer must be in 'read from' mode when compact() is called and will be in 'write to' mode
-             * afterwards.
-             */
+            //compact the buffer, this is an optional method, wonder what would happen if we didn't
             netInBuffer.compact();
             //read in the status
             handshakeStatus = result.getHandshakeStatus();
@@ -691,13 +690,11 @@ public class SecureNioChannel extends NioChannel {
         int read = 0;
         //the SSL engine result
         SSLEngineResult unwrap;
-        boolean processOverflow = false;
+        OverflowState overflowState = OverflowState.NONE;
         do {
-            boolean useOverflow = false;
-            if (processOverflow) {
-                useOverflow = true;
+            if (overflowState == OverflowState.PROCESSING) {
+                overflowState = OverflowState.DONE;
             }
-            processOverflow = false;
             //prepare the buffer
             netInBuffer.flip();
             //unwrap the data
@@ -708,7 +705,7 @@ public class SecureNioChannel extends NioChannel {
             if (unwrap.getStatus() == Status.OK || unwrap.getStatus() == Status.BUFFER_UNDERFLOW) {
                 //we did receive some data, add it to our total
                 read += unwrap.bytesProduced();
-                if (useOverflow) {
+                if (overflowState == OverflowState.DONE) {
                     // Remove the data read into the overflow buffer
                     read -= getBufHandler().getReadBuffer().position();
                 }
@@ -768,14 +765,15 @@ public class SecureNioChannel extends NioChannel {
                         dsts = dsts2;
                         length++;
                         getBufHandler().configureReadBufferForWrite();
-                        processOverflow = true;
+                        overflowState = OverflowState.PROCESSING;
                     }
                 }
             } else {
                 // Something else went wrong
                 throw new IOException(sm.getString("channel.nio.ssl.unwrapFail", unwrap.getStatus()));
             }
-        } while (netInBuffer.position() != 0 || processOverflow); //continue to unwrapping as long as the input buffer has stuff
+        } while ((netInBuffer.position() != 0 || overflowState == OverflowState.PROCESSING) &&
+                overflowState != OverflowState.DONE);
         return read;
     }
 
@@ -897,4 +895,11 @@ public class SecureNioChannel extends NioChannel {
     public ByteBuffer getEmptyBuf() {
         return emptyBuf;
     }
+
+
+    private enum OverflowState {
+        NONE,
+        PROCESSING,
+        DONE;
+    }
 }
