@@ -39,6 +39,7 @@ import javax.servlet.http.HttpServletRequest;
 import javax.servlet.http.HttpServletResponse;
 
 import org.junit.Assert;
+import org.junit.Assume;
 import org.junit.Test;
 
 import org.apache.catalina.Context;
@@ -86,20 +87,20 @@ public class TestSsl extends TomcatBaseTest {
 
     @Test
     public void testPost() throws Exception {
-        final SocketFactory socketFactory = TesterSupport.configureClientSsl();
+        SocketFactory socketFactory = TesterSupport.configureClientSsl();
 
         Tomcat tomcat = getTomcatInstance();
         TesterSupport.initSsl(tomcat);
         // Increase timeout as default (3s) can be too low for some CI systems
         Assert.assertTrue(tomcat.getConnector().setProperty("connectionTimeout", "20000"));
 
-        Context ctxt = getProgrammaticRootContext();
+        Context ctxt = tomcat.addContext("", null);
         Tomcat.addServlet(ctxt, "post", new SimplePostServlet());
         ctxt.addServletMappingDecoded("/post", "post");
         tomcat.start();
         int iterations = 8;
-        final CountDownLatch latch = new CountDownLatch(iterations);
-        final AtomicInteger errorCount = new AtomicInteger(0);
+        CountDownLatch latch = new CountDownLatch(iterations);
+        AtomicInteger errorCount = new AtomicInteger(0);
         for (int i = 0; i < iterations; i++) {
             new Thread() {
                 @Override
@@ -172,7 +173,7 @@ public class TestSsl extends TomcatBaseTest {
         ctxt.addApplicationListener(WsContextListener.class.getName());
 
         TesterSupport.initSsl(tomcat, TesterSupport.LOCALHOST_KEYPASS_JKS,
-                TesterSupport.JKS_PASS, null, TesterSupport.JKS_KEY_PASS, null);
+                TesterSupport.JKS_PASS, TesterSupport.JKS_KEY_PASS);
 
         tomcat.start();
         ByteChunk res = getUrl("https://localhost:" + getPort() +
@@ -182,40 +183,22 @@ public class TestSsl extends TomcatBaseTest {
                 TesterSupport.getLastClientAuthRequestedIssuerCount() == 0);
     }
 
-    @Test
-    public void testKeyPassFile() throws Exception {
-        TesterSupport.configureClientSsl();
-
-        Tomcat tomcat = getTomcatInstance();
-
-        File appDir = new File(getBuildDirectory(), "webapps/examples");
-        Context ctxt  = tomcat.addWebapp(null, "/examples", appDir.getAbsolutePath());
-        ctxt.addApplicationListener(WsContextListener.class.getName());
-
-        TesterSupport.initSsl(tomcat, TesterSupport.LOCALHOST_KEYPASS_JKS,
-                      null, TesterSupport.JKS_PASS_FILE, null, TesterSupport.JKS_KEY_PASS_FILE);
-
-        tomcat.start();
-        ByteChunk res = getUrl("https://localhost:" + getPort() +
-            "/examples/servlets/servlet/HelloWorldExample");
-        Assert.assertTrue(res.toString().indexOf("<a href=\"../helloworld.html\">") > 0);
-        Assert.assertTrue("Checking no client issuer has been requested",
-                TesterSupport.getLastClientAuthRequestedIssuerCount() == 0);
-    }
 
     @Test
-    public void testClientInitiatedRenegotiation() throws Exception {
-
+    public void testRenegotiateWorks() throws Exception {
         Tomcat tomcat = getTomcatInstance();
-        TesterSupport.initSsl(tomcat);
 
-        boolean renegotiationSupported = TesterSupport.isClientRenegotiationSupported(getTomcatInstance());
+        Assume.assumeTrue("SSL renegotiation has to be supported for this test",
+                TesterSupport.isClientRenegotiationSupported(getTomcatInstance()));
 
         Context root = tomcat.addContext("", TEMP_DIR);
-        Wrapper w = Tomcat.addServlet(root, "tester", new TesterServlet());
+        Wrapper w =
+            Tomcat.addServlet(root, "tester", new TesterServlet());
         w.setAsyncSupported(true);
         root.addServletMappingDecoded("/", "tester");
 
+        TesterSupport.initSsl(tomcat);
+
         tomcat.start();
 
         SSLContext sslCtx;
@@ -229,7 +212,8 @@ public class TestSsl extends TomcatBaseTest {
         }
         sslCtx.init(null, TesterSupport.getTrustManagers(), null);
         SSLSocketFactory socketFactory = sslCtx.getSocketFactory();
-        SSLSocket socket = (SSLSocket) socketFactory.createSocket("localhost", getPort());
+        SSLSocket socket = (SSLSocket) socketFactory.createSocket("localhost",
+                getPort());
 
         OutputStream os = socket.getOutputStream();
         InputStream is = socket.getInputStream();
@@ -244,17 +228,7 @@ public class TestSsl extends TomcatBaseTest {
 
         socket.startHandshake();
 
-        try {
-            doRequest(os, r);
-            if (!renegotiationSupported) {
-                Assert.fail("Renegotiation started when it should have failed");
-            }
-        } catch (IOException e) {
-            if (renegotiationSupported) {
-                Assert.fail("Renegotiation failed when it should be supported");
-            }
-            return;
-        }
+        doRequest(os, r);
         // Handshake complete appears to be called asynchronously
         int wait = 0;
         while (wait < 5000 && !listener.isComplete()) {
